package org.bolson.vote;
import java.io.*;

/**
 * Command line app for processing votes.
 */
public class vote {
    boolean rating = true;
    boolean isFloat = false;
    boolean isBin = false;
    boolean scriptMode = false;
    boolean htmlOut = false;
    boolean printState = false;
    boolean quiet = false;
    boolean printVotes = false;
    boolean validateVotes = false;
    boolean historgam = false;
    Class fac = null;
	String[] facArgv = new String[0];
    int numc = -1;
    boolean histogram = false;
    int histBuckets = -1;
    String names[] = null;
    int histyle = 2;
    VotingSystem tab = null;
    Histogram hist = null;
    int votes = 0;
	
	static abstract class VoteSource {
		abstract int[] getIntVote( int[] pi ) throws java.io.IOException;
		abstract float[] getFloatVote( float[] pf ) throws java.io.IOException;
		abstract double[] getDoubleVote( double[] pf ) throws java.io.IOException;
		public int numc;
	}
	static class BinVoteSource extends VoteSource {
		public DataInputStream dis = null;
		public BinVoteSource( DataInputStream din ) {
			dis = din;
		}
		int[] getIntVote( int[] pi ) throws java.io.IOException {
			for ( int i = 0; i < numc; i++ ) {
				pi[i] = dis.readInt();
			}
			return pi;
		}
		public float[] getFloatVote( float[] pf ) throws java.io.IOException {
			for ( int i = 0; i < numc; i++ ) {
				pf[i] = dis.readFloat();
			}
			return pf;
		}
		public double[] getDoubleVote( double[] pf ) throws java.io.IOException {
			for ( int i = 0; i < numc; i++ ) {
				pf[i] = dis.readDouble();
			}
			return pf;
		}
	}
	static class TextVoteSource extends VoteSource {
		public StreamTokenizer st = null;
		public TextVoteSource( StreamTokenizer stin ) {
			st = stin;
		}
		public int[] getIntVote( int[] pf ) throws java.io.IOException {
			int tt;
			for ( int i = 0; i < numc; i++ ) {
				tt = st.nextToken();
				if ( tt == st.TT_EOL ) {
					if ( i != 0 ) {
						System.err.println("\twrong number of votes on a line. eol at i="+i );
						return null;
					} else {
						tt = st.nextToken();
					}
				}
				if ( tt != st.TT_NUMBER ) {
					return null;
				}
				pf[i] = (int)st.nval;
			}
			return pf;
		}
		public float[] getFloatVote( float[] pf ) throws java.io.IOException {
			int tt;
			for ( int i = 0; i < numc; i++ ) {
				tt = st.nextToken();
				if ( tt == st.TT_EOL ) {
					if ( i != 0 ) {
						System.err.println("\twrong number of votes on a line. eol at i="+i );
						return null;
					} else {
						tt = st.nextToken();
					}
				}
				if ( tt != st.TT_NUMBER ) {
					return null;
				}
				pf[i] = (float)st.nval;
			}
			return pf;
		}
		public double[] getDoubleVote( double[] pf ) throws java.io.IOException {
			int tt;
			for ( int i = 0; i < numc; i++ ) {
				tt = st.nextToken();
				if ( tt == st.TT_EOL ) {
					if ( i != 0 ) {
						System.err.println("\twrong number of votes on a line. eol at i="+i );
						return null;
					} else {
						tt = st.nextToken();
					}
				}
				if ( tt != st.TT_NUMBER ) {
					return null;
				}
				pf[i] = st.nval;
			}
			return pf;
		}
	}
    
	public void initHistogram() {
		int nbuck = histBuckets;
		double min, max;
		if ( nbuck == -1 ) {
			if ( rating ) {
				nbuck = 11;
			} else {
				nbuck = numc;
			}
		}
		if ( rating ) {
			if ( isFloat ) {
				min = -1.0;
				max = 1.0;
			} else {
				min = 0;
				max = 10;
			}
		} else {
			min = 1;
			max = numc;
		}
		hist = new Histogram( numc, nbuck, min, max );
    }
	public void vote( InputStream is ) {
		if ( ! isBin ) {
			vote( new InputStreamReader( is ) );
			return;
		}
		DataInputStream dis = new DataInputStream( is );
		vote( dis );
	}
	public void vote( DataInputStream dis ) {
		VoteSource vs = new BinVoteSource( dis );
		vs.numc = numc;
		voteFromVoteSourceWithInit( vs );
	}
	public void vote( Reader r ) {
		if ( isBin ) {
			System.err.println( "vote(Reader) only for text voting" );
			return;
		}
		StreamTokenizer st = new StreamTokenizer( r );
		VoteSource vs = new TextVoteSource( st );
		vs.numc = numc;
		voteFromVoteSourceWithInit( vs );
	}
public void voteFromTextFile( String filename ) throws Exception {
    StreamTokenizer st = null;
    DataInputStream dis = null;
    boolean goodP = true;
	VoteSource vs;

    if ( validateVotes && isFloat && ! rating ) {
	System.err.println("floating point rankings are nonsensical. (fractional placement?) aborting. not processing file \"" + filename + '"' );
	return;
    }
//    System.out.println("reading votes from \"" + filename + "\", processing with " + fac + ", " + numc + " choices, " + (rating ? "Rating " : "Ranking " ) + (isFloat ? "float " : "int ") + (isBin ? "bin" : "text") );
    if ( isBin ) {
		dis = new DataInputStream( new FileInputStream( filename ));
		if ( numc == -1 ) {
			numc = dis.readInt();
		}
		vs = new BinVoteSource( dis );
    } else {
	st = new StreamTokenizer( new FileReader( filename ));
	st.eolIsSignificant( true );
	if ( numc == -1 ) {
	    if ( st.nextToken() != st.TT_NUMBER ) {
		System.out.println( filename + ": numc undefined and first token of file not a number");
		return;
	    }
	    numc = (int)st.nval;
	    if ( st.nextToken() != st.TT_EOL ) {
		System.out.println( filename + ": numc undefined and first line not a number by itself");
	    }
	}
	vs = new TextVoteSource( st );
    }
	vs.numc = numc;
	voteFromVoteSourceWithInit( vs );
}
public void voteFromVoteSourceWithInit( VoteSource vs ) {
	if ( tab == null ) {
		try {
			tab = (VotingSystem)(fac.getConstructors())[0].newInstance( new Object[]{new Integer( numc )} );
			tab.init( facArgv );
		} catch ( Exception e ) {
			e.printStackTrace();
			return;
		}
	}
    if ( histogram && hist == null ) {
		initHistogram();
	}
	voteFromVoteSource( vs );
}
public void printVote( float pf[] ) {
	for ( int i = 0; i < pf.length; i++ ) {
		if ( i != 0 ) {
			System.out.print( '\t' );
		}
		System.out.print( pf[i] );
	}
	System.out.println();
}
public void printVote( int pf[] ) {
	for ( int i = 0; i < pf.length; i++ ) {
		if ( i != 0 ) {
			System.out.print( '\t' );
		}
		System.out.print( pf[i] );
	}
	System.out.println();
}
public void vote( float pf[] ) {
	if ( rating ) {
		tab.voteRating( pf );
		if ( hist != null ) {
			hist.addRating( pf );
		}
	} else {
		/* this is gross, never do this */
		int pi[] = new int[numc];
		for ( int i = 0; i < numc; i++ ) {
			pi[i] = (int)pf[i];
		}
		tab.voteRanking( pi );
		if ( hist != null ) {
			hist.addRanking( pi );
		}
	}
}
public void vote( int pi[] ) {
	if ( rating ) {
		tab.voteRating( pi );
		if ( hist != null ) {
			hist.addRating( pi );
		}
	} else {
		tab.voteRanking( pi );
		if ( hist != null ) {
			hist.addRanking( pi );
		}
	}
}
public void voteFloatFromVoteSource( VoteSource vs ) {
    float pf[] = null;
	pf = new float[numc];
    try { while ( true ) {
		//	System.out.print("vote:");
		pf = vs.getFloatVote( pf );
		if ( pf == null ) break;
	    if ( printVotes ) {
			printVote( pf );
		}
		vote( pf );
	    votes++;
    }} catch ( Exception e ) {
		e.printStackTrace();
    };
}
public void voteIntFromVoteSource( VoteSource vs ) {
    int pf[] = null;
	pf = new int[numc];
    try { while ( true ) {
		//	System.out.print("vote:");
		pf = vs.getIntVote( pf );
		if ( pf == null ) break;
	    if ( printVotes ) {
			printVote( pf );
		}
		vote( pf );
	    votes++;
    }} catch ( Exception e ) {
		e.printStackTrace();
    };
}
public void voteFromVoteSource( VoteSource vs ) {
    if ( isFloat ) {
		voteFloatFromVoteSource( vs );
    } else {
		voteIntFromVoteSource( vs );
    }
}
/*public void voteFromVoteSource( VoteSource vs ) {
    int pi[] = null;
    float pf[] = null;
    if ( isFloat ) {
		pf = new float[numc];
    } else {
		pi = new int[numc];
    }
    try { while ( true ) {
		//	System.out.print("vote:");
		if ( isFloat ) {
		    pf = vs.getFloatVote( pf );
			if ( pf == null ) break;
		} else {
		    pi = vs.getIntVote( pi );
			if ( pi == null ) break;
		}
	    if ( printVotes ) {
			for ( int i = 0; i < numc; i++ ) {
				if ( i != 0 ) {
					System.out.print( '\t' );
				}
				if ( isFloat ) {
					System.out.print( pf[i] );
				} else {
					System.out.print( pi[i] );
				}
			}
			System.out.println();
		}
	    if ( isFloat ) {
			vote( pf );
	    } else {
			vote( pi );
	    }
	    votes++;
    }} catch ( Exception e ) {
		e.printStackTrace();
    };
}*/
public void displayTab() {
    //System.out.println( votes + " votes counted" );
    int[] winners = tab.getWinners();
    if ( printState ) {
	if ( names != null ) {
	    System.out.print( tab.toString( names ) );
	} else {
	    System.out.print( tab.toString() );
	}
    }
    if ( htmlOut ) {
	System.out.print( tab.htmlSummary( names ) );
    }
    if ( scriptMode ) {
	for ( int i = 0; i < winners.length; i++ ) {
	    System.out.println( winners[i] );
	}
    } else if ( winners.length == 0 ) {
	System.out.println("no winner found,"+tab);
    } else if ( winners.length == 1 ) {
	if ( names != null ) {
	    System.out.println( "Winner: " + names[winners[0]] );
	} else {
	    System.out.println("winner, choice number " + (winners[0] + 1) + " (of 1.." + numc + ")" );
	}
    } else {
	if ( names != null ) {
	    System.out.print( "Winners: " + names[winners[0]] );
	    for ( int i = 1; i < winners.length; i++ ) {
		System.out.print(", " + names[winners[i]] );
	    }
	} else {
	    System.out.print( winners.length + " winners, numbers " + (winners[0] + 1) );
	    for ( int i = 1; i < winners.length; i++ ) {
		System.out.print(", " + (winners[i] + 1) );
	    }
	}
	System.out.println(" (of 1.." + numc + ")" );
    }
    if ( hist != null ) {
	System.out.print( hist.toString( histyle, names ) );
    }
}

static final String optstring = "bBcdDfhi:n:Rrw";
static final String usageText = "vote [-bBcdfhRrw][-n number of choices][-i vote file][vote file ...]\n";
static final String helpText =
"\t-b\tbinary\n\t-B\ttext (default)\n\t-c\tCondorcet\n\t-D\tBorda count (default)\n\t-d\tinteger (default)\n\t-f\tfloat (-1.0 .. 1.0)\n\t-h\tthis help\n\t-H\toutput html table\n\t-i f\timmediately load and run vote file f\n\t-I\tInstant Runoff Vote\n\t-n N\tnumber of choices\n\t-N f\tfile with names of choices\n\t-p\tprint full state, not just winner\n\t-q\tquiet, don't print winner\n\t-R\tnumbers are rankings (1 best) (default)\n\t-r\tnumbers are ratings (higher better)\n\t-s\tscript friendly output (zero indexed \\n separated list of winners)\n\t-S\tmake histogram of ranked or rated votes\n\t-T #\thistogram style #\n\t-w\tRaw Sum of ratings\n\t-Y\tverify that votes are valid\n";

public static void main( String argv[] ) {
    vote it = new vote();
    argv = it.readArgv( argv );
    if ( argv != null ) for ( int i = 0; i < argv.length; i++ ) {
	try {
	    it.voteFromTextFile( argv[i] );
		it.displayTab();
	} catch ( Exception e ) {
	    e.printStackTrace();
	}
    }
}
/**
 * @param argv like main()
 * @return unused arguments (filenames)
 */
public String[] readArgv( String argv[] ) {
    int c;
    int i;
	java.util.Vector vFacArgv = new java.util.Vector();

    try {
	fac = Class.forName("org.bolson.vote.RawRating");
    } catch ( java.lang.ClassNotFoundException e ) {
	e.printStackTrace();
    }
    for ( i = 0; i < argv.length; i++ ) {
	if ( argv[i].equals("-B") || argv[i].equals( "--text" ) ) {
	    isBin = false;
	} else if ( argv[i].equals("-b") || argv[i].equals( "--binary" ) ) {
	    isBin = true;
	} else if ( argv[i].equals("-c") || argv[i].equals( "--condorcet" ) ) {
	    try {
		fac = Class.forName("org.bolson.vote.Condorcet");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-D") || argv[i].equals( "--borda" ) ) {
	    try {
		fac = Class.forName("org.bolson.vote.BordaVotingSystem");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-d") ) {
	    isFloat = false;
	} else if ( argv[i].equals("-f") ) {
	    isFloat = true;
	} else if ( argv[i].equals("-h") ) {	// help
	    System.out.println(usageText);
	    System.out.println(helpText);
	} else if ( argv[i].equals("-H") ) {
	    htmlOut = true;
	} else if ( argv[i].equals("-I") || argv[i].equals( "--IRV" ) ) {
	    try {
		fac = Class.forName("org.bolson.vote.IRV");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-i") ) {
	    i++;
	    try {
		voteFromTextFile( argv[i] );
	    } catch ( Exception e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-M") ) {
		i++;
		vFacArgv.add( argv[i] );
	} else if ( argv[i].equals("-N") ) {
		String[] tnam;
	    i++;
	    tnam = readNames( argv[i] );
		if ( tnam != null ) {
			setNames( names );
		}
	} else if ( argv[i].equals("-n") ) {
	    i++;
	    numc = Integer.parseInt( argv[i] );
	} else if ( argv[i].equals("-p") ) {
	    printState = true;
	} else if ( argv[i].equals("-q") ) {
	    quiet = true;
	} else if ( argv[i].equals("-R") ) {
	    rating = false;
	} else if ( argv[i].equals("-r") ) {
	    rating = true;
	} else if ( argv[i].equals("-s") ) {
	    scriptMode = true;
	} else if ( argv[i].equals("-S") ) {
	    histogram = true;
	} else if ( argv[i].equals("-T") ) {
	    i++;
	    histyle = Integer.parseInt( argv[i] );
	    histogram = true;
	} else if ( argv[i].equals("-w") || argv[i].equals( "--raw" ) ) {
	    try {
		fac = Class.forName("org.bolson.vote.RawRating");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-W") || argv[i].equals( "--IRNR" )) {
	    try {
			fac = Class.forName("org.bolson.vote.IRNR");
	    } catch ( java.lang.ClassNotFoundException e ) {
			e.printStackTrace();
	    }
	} else if ( argv[i].equals("-Y") ) {
	    validateVotes = true;
	} else if ( argv[i].equals("--system") ) {
	    i++;
	    try {
		fac = Class.forName(argv[i]);
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("--printVotes") ) {
	    printVotes = true;
	} else if ( argv[i].equals("--") ) {
	    break;
	} else if ( argv[i].startsWith("-") ) {
	    System.err.println("bogus flag: " + argv[i] );
	    return null;
	} else {
	    break;
	}
    }
	if ( vFacArgv.size() != 0 ) {
		facArgv = new String[vFacArgv.size()];
		for ( int j = 0; j < vFacArgv.size(); j++ ) {
			facArgv[j] = (String)vFacArgv.elementAt( j );
		}
	}
    String toret[] = new String[argv.length - i];
    for ( int j = 0; i + j < argv.length; j++ ) {
	toret[j] = argv[i+j];
    }
    return toret;
}

public void setNames( String[] namesi ) {
	names = namesi;
	if ( numc == -1 ) {
		numc = names.length;
	}	
}
/**
 * @param filename A file with the names of choices, one per line.
 * @return The names (lines) of the file.
 */
static String[] readNames( String filename ) {
    try {
	FileReader fr = new FileReader( filename );
	BufferedReader br = new BufferedReader( fr );
	java.util.Vector tn = new java.util.Vector();
	String s;
	while ( (s = br.readLine()) != null ) {
	    tn.addElement( s );
	}
	br.close();
	String[] toret = new String[tn.size()];
	for ( int i = 0; i < toret.length; i++ ) {
	    toret[i] = (String)tn.elementAt( i );
	}
	return toret;
    } catch ( Exception e ) {
	return null;
    }
}
}
